"""ttcplib

A couple of utility classes useful for controlling the execution of the
ADI-TTCP network throughput test program and for obtaining the results.

Note that these classes are probably only suitable for use in command line
scripts, not GUI-based ones. The use of select.select() and Telnet.read_until()
don't fit well with a GUI program's event loop. See ttcpcontrol.py for a more
GUI-friendly implementation.
"""

import sys
import time
import telnetlib
import select
import re

# protocol identifiers
TCP = 0
UDP = 1

# the default TCP port on which ADI-TTCP listens for control connections
DEFAULT_CONTROL_PORT = 6000

# default values for Transfer attributes
DEFAULT_DATA_PORT    = 7000     # receive port (TCP and UDP); can only be used
                                # once per protocol per batch of Transfers
DEFAULT_PROTOCOL     = TCP      # the protocol to use
DEFAULT_BUFLEN       = 1000     # the size of buffer to write or read
DEFAULT_NUMBUFS      = 20000    # the number of buffers to send
DEFAULT_DELAY        = 0        # microseconds of delay between writes
DEFAULT_RATE         = 0        # desired transmit rate in bytes/sec
DEFAULT_PERIOD       = 0        # number of seconds for which to write
DEFAULT_CONTINUOUS   = False    #
DEFAULT_TCP_NODELAY  = False    # set TCP_NODELAY socket option?
DEFAULT_SO_DEBUG     = False    # set SO_DEBUG socket option?
DEFAULT_SO_SNDBUF    = 0        # set SO_SNDBUF socket option?
DEFAULT_SO_RCVBUF    = 0        # set SO_RCVBUF socket option?
DEFAULT_ALIGN        = 0        # buffer alignment
DEFAULT_OFFSET       = 0        # buffer offset
DEFAULT_FORMAT       = 'K'      # throughput reporting unit
DEFAULT_TOUCH        = False    # touch received data?

#-------------------------------------------------------------------------------

class Controller(telnetlib.Telnet):

    """ADI-TTCP Controller class.

    Controller extends the standard Telnet client class with the addition
    of an optional description string in the constructor and a method for
    obtaining the host, port and description of an instance.

    A script should create one instance of Controller for each instance of the
    ADI-TTCP program with which it wishes to communicate. It should use the
    Telnet methods, such as write() and expect(), to perform the communication.

    See the telnetlib documentation for further information.
    """

    def __init__(self, host=None, port=DEFAULT_CONTROL_PORT, descr=""):
        """Constructor.

        Calls the Telnet constructor with the 'host' and 'port' arguments. If
        'host' is None then the Controller instance remains unconnected and
        the (Telnet) method open() must be used to open a connection; otherwise
        an attempt is made to connect to the ADI-TTCP program that is awaiting
        a control connection on 'port' of 'host'.

        Optional argument 'descr' is a string that is intended to describe the
        platform on which the controlled ADI-TTCP program is running, perhaps
        naming the CPU type and giving relevant clock speeds, etc.
        """
        telnetlib.Telnet.__init__(self, host, port)
        self.host  = host
        self.port  = port
        self.descr = descr

    def getControlInfo(self):
        """ Return information about this Control instance.

        Returns a tuple of the host, port and description associated with
        this instance of Controller.
        """
        return (self.host, self.port, self.descr)

#-------------------------------------------------------------------------------

class Transfer(object):

    """ADI-TTCP Transfer class.

    A Transfer instance holds sufficient information about a throughput test
    to allow it to issue appropriate TRANSFER commands to the two instances
    of the ADI-TTCP program involved in the test.

    A script should create one instance of Transfer for each throughput test
    that is to be run concurrently and then call static function Run() to
    execute the tests and obtain the results.

    The Transfer constructor requires two positional arguments, instances of
    Controller for the transmit and receive end respectively, and accepts
    any of the following optional keyword arguments:

        port       = integer: port number for receive end of transfer
        proto      = integer: ttcplib.TCP or ttcplib.UDP
        buflen     = integer: size of writes to and reads from socket
        numbufs    = integer: number of buffers to write
        delay      = integer: microseconds of delay to insert between writes
        rate       = integer: desired transfer rate (in bytes per second)
        period     = integer: number of seconds for which to transmit
        continuous =
        nodelay    = boolean: whether or not to set TCP_NODELAY socket option
        debug      = boolean: whether or not to set SO_DEBUG socket option
        sndbuf     = integer: size for SO_SNDBUF socket option
        rcvbuf     = integer: size for SO_RCVBUF socket option
        align      = integer: alignment for send and receive buffers
        offset     = integer: offset within buffer at which to send/receive
        format     = string:  unit for throughput message ('M', 'K', etc)
        touch      = boolean: whether or not to 'touch' each received byte

    """

    batch = list()  # list of Transfers to be processed by Run()

    def __init__(self,
                 ctrl1,     # transmitter's Controller instance
                 ctrl2,     # receiver's Controller instance
                 **kwargs):

        # set up default transfer attributes
        self.ctrl1      = ctrl1
        self.ctrl2      = ctrl2
        self.port       = DEFAULT_DATA_PORT
        self.proto      = DEFAULT_PROTOCOL
        self.buflen     = DEFAULT_BUFLEN
        self.numbufs    = DEFAULT_NUMBUFS
        self.delay      = DEFAULT_DELAY
        self.rate       = DEFAULT_RATE
        self.period     = DEFAULT_PERIOD
        self.continuous = DEFAULT_CONTINUOUS
        self.nodelay    = DEFAULT_TCP_NODELAY
        self.debug      = DEFAULT_SO_DEBUG
        self.sndbuf     = DEFAULT_SO_SNDBUF
        self.rcvbuf     = DEFAULT_SO_RCVBUF
        self.align      = DEFAULT_ALIGN
        self.offset     = DEFAULT_OFFSET
        self.format     = DEFAULT_FORMAT
        self.touch      = DEFAULT_TOUCH

        # now update attributes from the keyword arguments (if any)
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                print "Warning: unrecognised transfer attribute '" + \
                      key + \
                      "' ignored"

        # initialise runtime attributes
        self.xmtID          = 0         # TRANSFER -t identifier
        self.xmtTxt         = ""        # ... and result string
        self.rcvID          = 0         # TRANSFER -r identifier
        self.rcvTxt         = ""        # ... and result string

        # add this new Transfer to the current batch
        Transfer.batch.append(self)


    def Run():
        """Run() static method

        Iterates through the current batch of Transfers constructing a transmit
        and receive TRANSFER command for each and sending them to appropriate
        ADI-TTCP control streams. Then issues RUN command to all control streams
        and waits for all the transfers to complete.

        Returns the list of Transfers in the batch. The xmtTxt and rcvTxt
        members of each Transfer are strings containing the output returned
        from the transmit and receive sides of the transfer.  The output strings
        have been stripped of their "START RESULTS" / "END RESULTS" markers.

        """
        # set of control instances used in this batch
        controllers = set()

        # dictionary mapping Controller and transfer ID to Transfer
        tmap = dict()

        # construct and send TRANSFER commands for each Transfer in batch
        for t in Transfer.batch:
            xmtHost, xmtPort, xmtDescr = t.ctrl1.getControlInfo()
            rcvHost, rcvPort, rcvDescr = t.ctrl2.getControlInfo()

            # transmit side
            cmd  = "TRANSFER -t -p %d " % t.port
            if t.proto == UDP:cmd += "-u "
            if t.buflen > 0:  cmd += "-l %d " % (t.buflen)
            if t.delay > 0:   cmd += "-w %d " % (t.delay)
            if t.rate > 0:    cmd += "-L %d " % (t.rate)
            if t.period > 0:  cmd += "-P %d " % (t.period)
            if t.continuous or t.period > 0:
                cmd += "-c "
            else:
                cmd += "-n %d " % (t.numbufs)
            if t.nodelay:     cmd += "-D "
            if t.debug:       cmd += "-d "
            if t.sndbuf > 0:  cmd += "-b %d " % (t.sndbuf)
            if t.align > 0:   cmd += "-A %d " % (t.align)
            if t.offset > 0:  cmd += "-O %d " % (t.offset)
            cmd += "-f %s %s\r\n" % (t.format, rcvHost)

            t.ctrl1.write(cmd)

            # receive side
            cmd  = "TRANSFER -r -p %d " % t.port
            if t.proto == UDP:cmd += "-u "
            if t.buflen > 0:  cmd += "-l %d " % (t.buflen)
            if t.nodelay:     cmd += "-D "
            if t.debug:       cmd += "-d "
            if t.rcvbuf > 0:  cmd += "-b %d " % (t.rcvbuf)
            if t.align > 0:   cmd += "-A %d " % (t.align)
            if t.offset > 0:  cmd += "-O %d " % (t.offset)
            if t.touch:       cmd += "-T "
            cmd += "-f %s\r\n" % (t.format)

            t.ctrl2.write(cmd)

            # obtain the TRANSFER identifiers for each side
            txt = t.ctrl1.read_until('\r\n', 5)
            t.xmtID = int(txt)
            txt = t.ctrl2.read_until('\r\n', 5)
            t.rcvID = int(txt)

            # add transmit and receive side Controllers to set of
            # unique Controllers used in this batch
            controllers.add(t.ctrl1)
            controllers.add(t.ctrl2)

            # add entries to Transfer map
            tmap[(t.ctrl1,t.xmtID)] = (t, 'T')
            tmap[(t.ctrl2,t.rcvID)] = (t, 'R')

        # send a RUN command to every Controller involved in this batch
        for c in controllers:
            c.write("RUN\r\n")

        # loop waiting for output from each end of each transfer
        numResults = len(Transfer.batch) * 2
        while numResults > 0:
            ready, x, y = select.select(controllers, [], [])
            for c in ready:
                txt = c.read_until('END RESULTS\r\n')
                id = int(txt[0:txt.index(' ')])
                s = txt.index('\r\n') + 2
                e = txt.rindex('\r\n', 0, len(txt)-3)
                txt = txt[s:e]
                t, dir = tmap[(c,id)]
                if dir == 'T':
                    t.xmtTxt = txt
                else:
                    t.rcvTxt = txt
                numResults -= 1


        tfrlist = Transfer.batch

        # empty the list of current transfers
        Transfer.batch = list()

        # return results
        return tfrlist

    # make Run() a static method
    Run = staticmethod(Run)

#-------------------------------------------------------------------------------
